/*
 * anti_csrf.c: Simple in-form session handling
 *
 * Note that this is a limited implementation specific for
 * Nagios' needs. It should be adaptable for a more generic
 * use, but that's not my itch to scratch.
 *
 * Copyright(c) Andreas Ericsson <ae@op5.se>
 * License: GPLv2 or whatever the Nagios code you're compiling
 *          this with happens to be licensed as.
 */
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ipc.h> /* for ftok() */
#include <fcntl.h>
#include <dirent.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>
#include <dirent.h>
#include "sha1.h"
#include "form_session.h"

char *ncgi_session_id = NULL;

struct ifs_session {
	const char *remote_user;
	const char *remote_addr;
	char token[41]; /* make space for nul */
	time_t updated;
	};

char *ifs_token_dir = NULL;

#ifndef PATH_MAX
# define PATH_MAX 4096
#endif

#define IFS_MAX_SESSION_DATA_SIZE 100
#define IFS_SESSION_DURATION (15 * 60) /* a session lasts 15 minutes */

typedef struct ifs_session ifs_session;
static struct ifs_session session;

static void *xmalloc(size_t size) {
	void *ptr;
	ptr = malloc(size);
	memset(ptr, '@', size);

	return ptr;
	}

/*
 * "in" must be at least 20 bytes long
 * "out" must be 41 chars long
 */
static void sha1_to_str(const unsigned char *in, char *out) {
	int i, x = 0;
	const char hex_chars[] = "0123456789abcdef";

	for(i = 0; i < 20; i++) {
		unsigned char hi, lo;
		hi = in[i] >> 4;
		lo = in[i] & 0x0f;
		out[x++] = hex_chars[hi];
		out[x++] = hex_chars[lo];
		}
	out[x] = 0;
	}

static void ifs_srand(void) {
	/* A very weak random seed. This should be replaced */
	srand((unsigned int)(ftok("/", 17) ^ time(NULL)));
	}

static int ifs_mkdir_recursive(const char *base_path) {
	char *slash, *path;
	int result;

	result = mkdir(base_path, 0700);
	if(result < 0 && errno == EEXIST)
		return 0;
	if(result < 0 && errno != ENOENT) {
		return -1;
		}

	path = strdup(base_path);
	if(!path)
		return -1;
	slash = strchr(path + 1, '/');
	while(slash) {
		if(!slash[1])
			break;
		*slash = '\0';
		result = mkdir(path, 0700);
		if(result < 0) {
			if(errno != EEXIST && errno != ENOENT) {
				free(path);
				return -1;
				}
			}
		*slash = '/';
		slash = strchr(slash + 1, '/');
		}

	free(path);
	return 0;
	}

#define IFS_TOKENDIR_SUFFIX ".ncgi-form-session-tokens"
static char *ifs_get_token_dir(void) {
	if(!ifs_token_dir) {
		const char *tmp;
		char *token_dir;
		int tmp_len, suffix_len;

		tmp = getenv("TMP");
		if(!tmp)
			tmp = getenv("TMPDIR");
		if(!tmp)
			tmp = "/tmp";

		tmp_len = strlen(tmp);
		suffix_len = sizeof(IFS_TOKENDIR_SUFFIX);
		token_dir = xmalloc(tmp_len + suffix_len + 2);
		memcpy(token_dir, tmp, tmp_len);
		token_dir[tmp_len] = '/';
		memcpy(token_dir + tmp_len + 1, IFS_TOKENDIR_SUFFIX, suffix_len);

		ifs_token_dir = strdup(token_dir);
		}

	if(ifs_mkdir_recursive(ifs_token_dir) < 0)
		return NULL;

	return ifs_token_dir;
	}

static char *ifs_get_path(const char *id) {
	int id_len, token_dir_len, path_len;
	char *path, *token_dir;

	token_dir = ifs_get_token_dir();
	if(!id || !token_dir)
		return NULL;

	id_len = strlen(id);
	if(id_len != 40)
		return NULL;

	token_dir_len = strlen(token_dir);
	path_len = id_len + token_dir_len + 2;
	path = calloc(path_len, 1);
	if(!path)
		return NULL;

	memcpy(path, token_dir, token_dir_len);
	path[token_dir_len] = '/';
	memcpy(&path[token_dir_len + 1], id, id_len);

	return path;
	}

static void ifs_purge(void) {
	char cwd[PATH_MAX], *token_dir;
	DIR *dirp;
	struct dirent *de;

	token_dir = ifs_get_token_dir();
	if(!token_dir)
		return;

	dirp = opendir(token_dir);
	if(!dirp)
		return;

	getcwd(cwd, sizeof(cwd));
	if(chdir(token_dir))
		return;

	while((de = readdir(dirp))) {
		struct stat st;

		if(strlen(de->d_name) != sizeof(session.token) - 1 ||
		        stat(de->d_name, &st) < 0 ||
		        st.st_size > IFS_MAX_SESSION_DATA_SIZE ||
		        !st.st_size ||
		        st.st_mtime < time(NULL) - IFS_SESSION_DURATION) {
			unlink(de->d_name);
			continue;
			}
		}

	chdir(cwd);
	}

static void ifs_print_session(ifs_session *sess) {
	printf("remote_user: %s\n", sess->remote_user);
	printf("remote_addr: %s\n", sess->remote_addr);
	printf("token      : %s\n", sess->token);
	printf("updated    : %lu\n", sess->updated);
	}

static int ifs_save_session(void) {
	char *path;
	FILE *fp;
	mode_t old_umask;

	ifs_purge();
	if(!session.updated || !session.remote_user || !session.remote_addr ||
	        strlen(session.token) != sizeof(session.token) - 1) {
		return -1;
		}

	path = ifs_get_path(session.token);
	if(!path)
		return -1;

	old_umask = umask(0077);
	fp = fopen(path, "w");
	umask(old_umask);
	free(path);
	if(!fp) {
		return -1;
		}

	fprintf(fp, "%s\n%s\n%s\n", session.remote_user, session.remote_addr,
	        session.token);
	fclose(fp);

	return 0;
	}


/*
 * Create the token from remote_user, remote_addr, the unix timestamp
 * and a random number. Since we're using a cryptographically secure
 * hash to create the token, a single random number suffices
 */
static const char *ifs_create_token(void) {
	ncgi_SHA_CTX ctx;
	unsigned char rawhash[20];
	int i;

	session.remote_user = getenv("REMOTE_USER");
	if(!session.remote_user)
		return NULL;

	session.remote_addr = getenv("REMOTE_ADDR");
	if(!session.remote_addr)
		return NULL;

	session.updated = time(NULL);

	ncgi_SHA1_Init(&ctx);
	ncgi_SHA1_Update(&ctx, session.remote_user, strlen(session.remote_user));
	ncgi_SHA1_Update(&ctx, session.remote_addr, strlen(session.remote_addr));
	ncgi_SHA1_Update(&ctx, &session.updated, sizeof(session.updated));

	/* add some random numbers to the mix */
	ifs_srand();
	for(i = 0; i < 10; i++) {
		int r;

		r = rand();
		ncgi_SHA1_Update(&ctx, &r, sizeof(r));
		}

	ncgi_SHA1_Final(rawhash, &ctx);
	sha1_to_str(rawhash, session.token);

	if(!ifs_save_session())
		return session.token;

	return NULL;
	}

static int ifs_parse_session(const char *path, ifs_session *sess) {
	int fd, rd;
	char buf[IFS_MAX_SESSION_DATA_SIZE], *p1, *p2;
	struct stat st;

	if(!path || !sess)
		return -1;

	memset(sess, 0, sizeof(ifs_session));

	if(stat(path, &st) < 0) {
		return -1;
		}

	if(st.st_size > IFS_MAX_SESSION_DATA_SIZE) {
		return -1;
		}

	fd = open(path, O_RDONLY);
	if(fd < 0) {
		return -1;
		}

	rd = read(fd, buf, st.st_size);
	close(fd);

	if(rd != st.st_size) {
		return -1;
		}

	/* dump trailing newlines */
	while(buf[--rd] == '\n')
		;
	buf[rd + 1] = '\0';

	/* clumsy, but quick to hack up */
	p1 = strchr(buf, '\n');
	if(!p1)
		return -1;
	*(p1++) = '\0';
	p2 = strchr(p1, '\n');
	if(!p2)
		return -1;
	*(p2++) = '\0';
	if(strlen(p2) != sizeof(sess->token) - 1)
		return -1;

	sess->remote_user = strdup(buf);
	sess->remote_addr = strdup(p1);
	memcpy(sess->token, p2, sizeof(sess->token) - 1);
	sess->token[sizeof(sess->token) - 1] = '\0';
	sess->updated = st.st_mtime;

	return 0;
	}

static int ifs_load_session(const char *id) {
	char *path;
	int result;
	ifs_session sess;

	ifs_purge();

	memset(&session, 0, sizeof(session));

	path = ifs_get_path(id);
	result = ifs_parse_session(path, &sess);
	free(path);
	if(result < 0)
		return -1;

	if(memcmp(id, sess.token, sizeof(session.token) - 1))
		return -1;

	memcpy(&session, &sess, sizeof(session));

	return 0;
	}

int ifs_has_valid_token(int strict) {
	if(session.updated)
		return 1;

	if(!ncgi_session_id || ifs_load_session(ncgi_session_id) < 0)
		return 0;

	if(strcmp(session.remote_user, getenv("REMOTE_USER")))
		return 0;

	if(strict && strcmp(session.remote_addr, getenv("REMOTE_ADDR")))
		return 0;

	return 1;
	}

/*
 * If the use already has a token, use that. Otherwise
 * create one and return that
 */
const char *ifs_get_token(void) {
	const char *token;

	if(ifs_has_valid_token(0))
		return session.token;

	token = ifs_create_token();
	if(token)
		token = strdup(token);

	return token;
	}

#ifdef ANTI_CSRF_TESTING
int main(int argc, char **argv) {
	const char *token;
	const char *static_token = "8fe1c9681f41f0fe9fd476421d908ce03e950bd4";

	token = ifs_get_token();
	printf("Created token: %s\n", token);

	printf("\n----\nLoading key: %s\n", token);
	ifs_load_session(token);
	ifs_print_session(&session);

	printf("\n----\nLoading key: %s\n", static_token);
	ifs_load_session(static_token);
	ifs_print_session(&session);

	printf("\n----\nLoading key: %s\n", token);
	ifs_load_session(token);
	ifs_print_session(&session);

	return 0;
	}
#endif
